#include "pogohash.h"

#include "pogohash.conf.45.0"

uc_engine *uc;
csh        cs;
uint8_t   *code_buffer;
int        initialized = 0;

const int HASHSEED = 0x61247FBF;

uint8_t neon[] = { 
                   0x4f, 0xf4, 0x70, 0x00,  // mov.w r0, #0xf00000
                   0x01, 0xee, 0x50, 0x0f,  // mcr p15, #0x0, r0, c1, c0, #0x2
                   0xbf, 0xf3, 0x6f, 0x8f,  // isb sy 
                   0x4f, 0xf0, 0x80, 0x43,  // mov.w r3, #0x40000000
                   0xe8, 0xee, 0x10, 0x3a   // vmsr fpexc, r3
                 }; 
uint8_t stub[] = { 
                   0x90, 0x47               // blx r2
                 };
uint32_t HashFunctionAddr = 0;

uc_hook hhc_ins;
void POGO_hook(uc_engine *uc, uint64_t address, uint32_t size, void *user_data)
{
#ifdef TRACE_DEBUG
  if (initialized == 1)
  {
    static uint32_t regs[16] = { 0 };
    static int      ins_counter;
    static int      address_lo;
    static int      address_hi;
//  static int      address_last = 0x2000;

    // instruction counter
    if (address == POGO_FUNC)
    {
      address_lo = address;
      address_hi = address;
      ins_counter = 1;
    }
    else
    {
      if (address < address_lo) address_lo = address;
      if (address > address_hi) address_hi = address;
      ins_counter++;
    }

//  if ((((uint64_t)address - (uint64_t)address_last) < 0) ||
//      (((uint64_t)address - (uint64_t)address_last) > 4))
//  {
//    printf("DETECT: BRANCH\n");
//  }

    printf("---\n");
    printf("address: 0x%08lx\n", address);
    printf("  ");
    {
      uint32_t reg;
      int      i;

      for (i=0; i<16; i++)
      {
        switch (i)
        {
          case 13: uc_reg_read(uc, UC_ARM_REG_SP, &reg); printf(" SP"); break;
          case 14: uc_reg_read(uc, UC_ARM_REG_LR, &reg); printf(" LR"); break;
          case 15: uc_reg_read(uc, UC_ARM_REG_PC, &reg); printf(" PC"); break;
          default: uc_reg_read(uc, UC_ARM_REG_R0+i, &reg); printf("R%02d", i);
                   break;
        }
        if (regs[i] != reg) printf("*"); else printf(":");
        printf(" 0x%08x ", reg); 
        regs[i] = reg;
  
        if (((i % 4) == 3) && (i != 15)) printf("\n  ");
      }
    }
    printf("\n");

    // display the instruction (hack)
    {
      cs_insn* insn;
      int      i, cnt;

      cnt = cs_disasm(cs, &code_buffer[address], 10, address, 1, &insn);
      if (cnt > 0)
      {
        printf("opcodes: ");
        for (i=0; i<insn->size; i++) printf("%02x ", insn->bytes[i]);
        for (i=insn->size; i<4; i++) printf("   ");
        printf(":: %s\t%s\n", insn->mnemonic, insn->op_str);

        cs_free(insn, cnt);
     }
    }

    // instruction counter
    if (address == POGO_FUNC_END)
    {
      printf("\n");
      printf("address_lo: 0x%08lx ", (uint64_t)address_lo);
      printf("address_hi: 0x%08lx ", (uint64_t)address_hi);
      printf("instructions: %d\n", ins_counter);
    }

//  address_last = address;
  }
#endif
}

uc_hook hhc_mem_rw;
bool POGO_hook_mem(uc_engine *uc, uc_mem_type type, uint64_t address, int size, int64_t value, void *user_data)
{
#ifdef TRACE_DEBUG
  // we ignore anything above 0xd0000000 (heap, stack)
  if (address < 0xd0000000) 
  {   
    switch (size)
    {
      case 4:
           if (type == UC_MEM_READ)
             printf("[mem-r] len:%d at 0x%08lx val: 0x%08x\n", 
                    size, address, *((uint32_t *)&code_buffer[address]));
           else
           if (type == UC_MEM_WRITE)
             printf("[mem-w] len:%d at 0x%08lx val: 0x%08x\n", 
                    size, address, (uint32_t)value);
           break;
    }
  }
#endif
  return false;
}

uc_err POGO_createmap(uint64_t address, uint32_t size)
{
  uint32_t chunk = address % 0x1000;
  uint64_t taddr = address - chunk;
  uint32_t tsize = ((0xFFF + chunk + size) / 0x1000) * 0x1000;

  uc_err err = uc_mem_map(uc, taddr, tsize, UC_PROT_ALL);
  if (err)
  {
    printf("Failed on uc_mem_map() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

uc_err POGO_memmap(uint64_t address, uint8_t* buffer, uint32_t size)
{
  uc_err err = uc_mem_write(uc, address, buffer, size);
  if (err)
  {
    printf("Failed on uc_mem_write() with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  return UC_ERR_OK;
}

int POGO_init()
{
  unsigned int sp;
  uc_err err;
  FILE* fp;
  long size, i;
  int exportdata, exportaddr, cnt;
 
  // dynamically load shared library
#ifdef DYNLOAD
  if (!uc_dyn_load(NULL, 0))
  {
    printf("Error dynamically loading shared library.\n");
    printf("Please check that unicorn.dll/unicorn.so is available as well as\n");
    printf("any other dependent dll/so files.\n");
    printf("The easiest way is to place them in the same directory as this app.\n");
    return 1;
  }
#endif

  // Initialize emulator in ARM 32bit thumb mode
  err = uc_open(UC_ARCH_ARM, UC_MODE_THUMB, &uc);
  if (err)
  {
    printf("Failed on uc_open() with error returned: %u\n", err);
    return err;
  }

  // hook all instructions by having @begin > @end
  uc_hook_add(uc, &hhc_ins, UC_HOOK_CODE, (void *)POGO_hook, NULL, (uint64_t)1, (uint64_t)0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }
  uc_hook_add(uc, &hhc_mem_rw, UC_HOOK_MEM_READ | UC_HOOK_MEM_WRITE, POGO_hook_mem, NULL, 1, 0);
  if (err)
  {
    printf("Failed on uc_hook_add(code) with error returned: %u: %s\n", err, uc_strerror(err));
    return err;
  }

  // heap and stack allocation
  POGO_createmap(0xE0000000, 0x2000);
  sp = 0xE0000000 + 0x2000; // stack grows down within heap

  // stack
  uc_reg_write(uc, UC_ARM_REG_SP, &sp);

  // binary mapping
  fp = fopen(POGO_FILENAME, "rb");
  size = POGO_BIN_SIZE; 
  code_buffer = malloc(size);
  fseek(fp, 0, SEEK_SET);
  fread(code_buffer, size, 1, fp);
  fclose(fp);

  // memory prepare
  POGO_createmap(POGO_BIN_OFFSET, POGO_BIN_MAX - POGO_BIN_OFFSET);

#if (POGO_PATCH_VER == 45)
  // specific for pokemon.45.0
  {
    char patch[2] = { 0x00, 0xbf }; // NOP 
    memmove(code_buffer + 0x01B1769C, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B1769E, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B17CD4, patch, sizeof(patch)); 
    memmove(code_buffer + 0x01B17CD6, patch, sizeof(patch)); 
  }
  // specific for pokemon.45.0: end
#endif

  // neon support enable 
  POGO_createmap(0x1000, 0x1000);
  POGO_memmap(0x1000, neon, sizeof(neon));
#ifdef TRACE_DEBUG
  memmove(code_buffer + 0x1000, neon, sizeof(neon)); 
#endif

  // start code calling hash func
  POGO_createmap(0x2000, 0x1000);
  POGO_memmap(0x2000, stub, sizeof(stub));
#ifdef TRACE_DEBUG
  memmove(code_buffer + 0x2000, stub, sizeof(stub)); 
#endif

  // export address fill with any valid address 
  exportdata = POGO_EXP_DATA;
  exportaddr = POGO_EXP_ADDR;
  cnt = 0;
  for (i = 0; i < POGO_EXP_CNT; ++i)
  {
    int value = *(int*)(code_buffer + exportaddr + (i * 4));
    if (value == 0)
    {
      memmove(code_buffer + exportaddr + (i * 4), &exportdata, 4); 
      cnt++;
    }
  }
  printf("%d patches\n", cnt);

  // all segment
  POGO_memmap(POGO_BIN_OFFSET, 
              code_buffer + POGO_BIN_OFFSET, 
              POGO_BIN_SIZE - POGO_BIN_OFFSET);

  // enable neon
  err = uc_emu_start(uc, 0x1000, 0x1000 + sizeof(neon), 0, 0);
  if (err)
  {
    printf("Failed on uc_emu_start() with error returned %u: %s\n",
      err, uc_strerror(err));
    return -1;
  }

#ifdef TRACE_DEBUG
  // initialize the disassembler
  {
    cs_err err = cs_open(CS_ARCH_ARM, CS_MODE_THUMB, &cs);
    if (err)
    {
      printf("Failed on cs_open() with error returned %u: %s\n",
        err, cs_strerror(err));
      return -1;
    }

    cs_option(cs, CS_OPT_DETAIL, CS_OPT_OFF);
  }  
#endif

  HashFunctionAddr = POGO_FUNC;
  initialized = 1;

  return 0;
}

void POGO_quit()
{
#ifdef TRACE_DEBUG
  cs_close(&cs);
#endif

  // free resources
  uc_close(uc);

  free(code_buffer);

  // dynamically free shared library
#ifdef DYNLOAD
  uc_dyn_free();
#endif
}

uint64_t Hash(uint8_t* buffer, int size)
{
  uint32_t r0 = 0xE0001000;  // buffer
  uint32_t r1 = size;        // size
  uint32_t r2 = HashFunctionAddr + 1;
  uint64_t ret;
  uc_err err;

  uc_reg_write(uc, UC_ARM_REG_R0, &r0);
  uc_reg_write(uc, UC_ARM_REG_R1, &r1);
  uc_reg_write(uc, UC_ARM_REG_R2, &r2);

  POGO_memmap(r0, buffer, size);

  // execute code
  err = uc_emu_start(uc, 0x2000, 0x2000 + sizeof(stub), 0, 0);
  if (err)
  {
    printf("Failed on uc_emu_start() with error returned %u: %s\n",
      err, uc_strerror(err));
    return -1;
  }

  uc_reg_read(uc, UC_ARM_REG_R0, &r0);
  uc_reg_read(uc, UC_ARM_REG_R1, &r1);

  ret = (unsigned int)r1;
  ret = (ret << 32) | (unsigned int)r0;
  return (int64_t)ret;
}

int64_t Hash64(char* buffer, int size, int salt)
{
  int64_t ret;
  uint8_t* newbuff = (uint8_t *)malloc(size + 4);
  newbuff[0] = (salt >> 24) & 0xFF;
  newbuff[1] = (salt >> 16) & 0xFF;
  newbuff[2] = (salt >> 8) & 0xFF;
  newbuff[3] = (salt >> 0) & 0xFF;
  memcpy(newbuff + 4, buffer, size);
  ret = Hash(newbuff, size + 4);
  free(newbuff);

  return ret;
}

int Hash64to32(int64_t hash)
{
  return (unsigned int)(hash >> 0) ^ (unsigned int)(hash >> 32);
}

int Hash32(char* buffer, int size)
{
  int64_t ret = Hash64(buffer, size, HASHSEED);
  return Hash64to32(ret);
}
